import build
import pygit2 as vcs
import shutil
import twine.settings
from functools import cached_property
from mkdocs.__main__ import cli
from twine.commands.upload import upload as twine_upload

from fmtr.tools.iterator_tools import IndexList

gh_deploy = cli.commands["gh-deploy"].callback



from fmtr.tools import environment_tools as env
from fmtr.tools import http_tools as http
from fmtr.tools.constants import Constants
from fmtr.tools.infrastructure_tools.project import Project
from fmtr.tools.inherit_tools import Inherit
from fmtr.tools.logging_tools import logger
from fmtr.tools.path_tools import Path


class Releaser(Inherit[Project]):
    """

    Manages the release process for a project.

    The release process consists of two main phases:

    1. Repository operations:
       - Fetch latest changes from remote
       - Increment version numbers in relevant files

    2. Distribution and publishing:
       - Create GitHub release (for both public and private repositories)
       - Build Python packages (wheel and sdist)
       - Upload to package indexes (private registry always, PyPI if configured)

    """

    @logger.instrument("Releasing {self.paths.name_ns}...")
    def run(self):
        self.repo.fetch()
        self.increment()
        self.repo.push()
        self.repo.fetch()

        if self.is_dockerhub:
            from fmtr.tools.infrastructure_tools.stack import ProductionPublic
            stack = self.stacks.cls[ProductionPublic]
            stack.build()
            stack.push()

        self.package()
        self.release()

    @property
    def message(self):
        return f"Release version {self.version}"

    @logger.instrument("Incrementing version numbers {self.repo.origin.url}...")
    def increment(self):
        """

        Increment version numbers in project files, create a new commit, and rebase the release branch.
        
        """
        repo = self.repo

        main_ref = repo.lookup_reference("refs/heads/main")  # todo raise if not main?
        parent = main_ref.peel(vcs.Commit)

        # Make sure we're building on main's HEAD (not whatever HEAD currently is).
        index = repo.index
        index.read_tree(parent.tree)

        # Apply incrementors (they edit files in the working tree), then stage results.
        for incrementor in self.incrementors:
            paths = incrementor.apply()
            if not paths:
                continue
            if not isinstance(paths, list):
                paths = [paths]

            for path in paths:
                rel = str(Path(path).relative_to(self.paths.repo))
                index.add(rel)

        index.write()
        tree = index.write_tree(repo)

        commit_id = repo.create_commit(
            main_ref.name,
            repo.default_signature,
            repo.default_signature,
            self.message,
            tree,
            [parent.id],
        )

        try:
            repo.create_tag(
                self.tag,
                commit_id,
                vcs.GIT_OBJECT_COMMIT,
                repo.default_signature,
                self.message,
            )
        except Exception as exception:
            logger.warning(f"Failed to create tag: {exception}")

        branch_name = "release"
        ref_name = f"refs/heads/{branch_name}"
        try:
            release_ref = repo.lookup_reference(ref_name)
        except KeyError:
            target = repo.head.peel(vcs.Commit)
            release_ref = repo.create_branch(branch_name, target)

        release_commit = release_ref.peel(vcs.Commit)
        if repo.merge_base(commit_id, release_commit.id) != release_commit.id:
            raise RuntimeError("release has diverged from main")
        release_ref.set_target(commit_id)

        return commit_id

    @cached_property
    def path(self):
        return Path.temp() / self.name

    @cached_property
    def token(self):
        return env.get(Constants.GITHUB_TOKEN_KEY)

    @cached_property
    def incrementors(self):
        return IndexList[Incrementor](
            [
                IncrementorVersion(self),
                IncrementorHomeAssistantAddon(self),
                IncrementorChangelog(self),
            ]
        )

    @cached_property
    def packagers(self):
        return [PackageWheel(self), PackageSourceDistribution(self)]

    @cached_property
    def releases(self):
        releases = [
            ReleaseGithub(self),
            ReleasePackageIndexPrivate(self),
            ReleaseDocumentation(self)

        ]

        if self.is_pypi:
            release = ReleasePackageIndexPublic(self)
            releases.append(release)

        return releases

    def release(self):
        for release in self.releases:
            release.release()


    def package(self):
        if self.path.exists():
            logger.warning(f"Package directory already exists: {self.path}. Will be removed.")
            shutil.rmtree(self.path)

        self.path.mkdir(parents=True)

        for packager in self.packagers:
            packager.package()


class Incrementor(Inherit[Releaser]):

    @property
    def cls(self):
        return self.__class__

    def apply(self) -> Path | None:
        raise NotImplementedError


class IncrementorVersion(Incrementor):

    @cached_property
    def path(self):
        return self.paths.version


    def apply(self) -> Path | list[Path] | None:
        path = self.paths.version
        logger.info(f'Incrementing version file "{self.path}" {Constants.ARROW_RIGHT} {self.version}...')
        path.write_text(str(self.version))
        project = self.inherit_root
        project.incremented = True
        return path


class IncrementorHomeAssistantAddon(Incrementor):

    @cached_property
    def path(self):
        return self.paths.ha_addon_config

    @logger.instrument('Incrementing Home Assistant Add-On version "{self.path}"...')
    def apply(self) -> Path | None:
        if not self.path.exists():
            logger.warning(f"Home Assistant Add-On config file not found: {self.path}. Skipping.")
            return None

        data = self.path.read_yaml()
        data['version'] = str(self.version)
        self.path.write_yaml(data)
        return self.path


class IncrementorChangelogSymlink(Incrementor):

    @property
    def src(self):
        raise NotImplementedError

    @property
    def dest(self):
        return self.paths.docs_changelog.with_stem(f'{self.version}')

    def apply(self) -> Path | list[Path] | None:
        if not self.dest.exists():
            logger.warning(f"Symlink dest not found: {self.dest}. Skipping.")
            return None

        dest = self.dest.relative_to(self.paths.repo)

        self.src.unlink(missing_ok=True)
        self.src.symlink_to(dest)

        return self.src


class IncrementorChangelog(IncrementorChangelogSymlink):

    @property
    def src(self):
        return self.paths.changelog


    @logger.instrument('Incrementing Changelog "{self.path}"...')
    def apply(self) -> Path | list[Path] | None:
        path = self.paths.docs_changelog
        if not path.exists():
            logger.warning(f"New changelog not found: {path}. Skipping.")
            return None

        logger.info(f"Version tagging Changelog: {path} {Constants.ARROW_RIGHT} {self.dest}")
        path.rename(self.dest)

        paths = [self.dest, super().apply()]
        return paths


class Packager(Inherit[Releaser]):
    """
    
    Base class for packaging operations.
    
    """
    TYPE = None

    def package(self):
        builder = build.ProjectBuilder(str(self.paths.repo))
        with logger.span(f'Building {self.TYPE} distribution...'):
            path = builder.build(self.TYPE, str(self.path))
            logger.info(f'Build complete: {path}')


class PackageWheel(Packager):
    """
    
    Package as Python wheel.
    
    """
    TYPE = 'wheel'


class PackageSourceDistribution(Packager):
    """
    
    Package as source distribution.
    
    """
    TYPE = 'sdist'


class Release(Inherit[Releaser]):
    """
    
    Base class for release operations.
    
    """
    def release(self):
        raise NotImplementedError


class ReleaseGithub(Release):
    """
    
    Release to GitHub.
    
    """

    @property
    def url(self):
        return f"https://github.com/{Constants.ORG_NAME}/{self.paths.name_ns}/compare/v{self.versions.old}...v{self.versions.new}"

    @property
    def body(self):
        path_changelog = self.incrementors.cls[IncrementorChangelog].dest
        if path_changelog.exists():
            return path_changelog.read_text()
        else:
            return f'**Full Changelog**: [{self.versions.old} {Constants.ARROW_RIGHT} {self.versions.new}]({self.url})'

    def release(self):
        url = f"https://api.github.com/repos/{self.org}/{self.paths.name_ns}/releases"
        name = f'Release {self.tag}'

        headers = {
            "Authorization": f"Bearer {self.token}",
            "Accept": "application/vnd.github+json"
        }



        payload = {
            "tag_name": self.tag,
            "name": name,
            "body": self.body,
            "draft": False,
            "prerelease": self.versions.is_pre
        }

        with logger.span(f'Creating release "{name}"...'):
            response = http.client.post(url, json=payload, headers=headers)

        response.raise_for_status()

        data = response.json()
        url = data['html_url']
        logger.info(f"Release created: {url}")


class ReleasePackageIndex(Release):
    """
    
    Base class for package index releases.
    
    """
    
    TOKEN_KEY = None
    URL = None
    USERNAME = None
    NAME = None

    @cached_property
    def token(self):
        return env.get(self.TOKEN_KEY)

    @cached_property
    def settings(self):
        return twine.settings.Settings(
            repository_name=self.NAME,
            repository_url=self.URL,
            username=self.USERNAME,
            password=self.token,
            non_interactive=True,
            verbose=True
        )

    def warn(self):
        pass

    def release(self):
        with logger.span(f'Uploading package to PyPI index ({self.URL}) as {self.USERNAME}...'):
            self.warn()
            twine_upload(self.settings, [f'{self.path}/*'])


class ReleasePackageIndexPrivate(ReleasePackageIndex):
    """
    
    Release to private package index.
    
    """
    TOKEN_KEY = Constants.PACKAGE_INDEX_PRIVATE_TOKEN_KEY
    URL = Constants.PACKAGE_INDEX_PRIVATE_URL
    USERNAME = Constants.ORG_NAME


class ReleasePackageIndexPublic(ReleasePackageIndex):
    """
    
    Release to public package index, namely PyPI.
    
    """
    TOKEN_KEY = Constants.PACKAGE_INDEX_PUBLIC_TOKEN_KEY
    URL = None
    USERNAME = '__token__'
    NAME = "pypi"

    def warn(self):
        logger.error(f'Project "{self.paths.name_ns}" is being pushed to a PUBLIC Package Index!')


class ReleaseDocumentation(Release):

    @property
    def data(self):
        import io

        from mkdocs.config import load_config
        from material.extensions.emoji import twemoji, to_svg

        return dict(
            config_file=io.StringIO(""),
            site_dir='site',
            docs_dir=str(self.paths.docs),

            site_name=self.paths.name_ns,
            theme={
                "name": "material",
                "features": [
                    "content.code.annotate",
                    "content.code.copy",
                ],
            },
            plugins=[
                "search",
                {"include_dir_to_nav": {"reverse_sort_file": True}},
                {
                    "mkdocstrings": {
                        "handlers": {
                            "python": {
                                "options": {
                                    "show_source": True,
                                }
                            }
                        }
                    }
                },
            ],
            markdown_extensions=[
                "admonition",
                "attr_list",
                "md_in_html",
                "pymdownx.superfences",
                {"pymdownx.highlight": {"pygments_lang_class": True}},
                {"pymdownx.snippets": {"check_paths": True}},

                {"pymdownx.tabbed": {"alternate_style": True}},
                {"pymdownx.emoji": {"emoji_index": twemoji, "emoji_generator": to_svg}},
            ],
            nav=self.nav or [
                {"Home": "index.md"},
                {"Changelog": "changelog/"},
            ],
            extra={
                "version": {"provider": "mike"},
            },
        )

    @property
    def message(self):
        return f"Release documentation version {self.version}"

    def deploy(self):
        result = gh_deploy(
            clean=True,
            message=self.message,
            remote_branch="docs",
            remote_name="origin",
            force=True,
            no_history=False,
            ignore_version=False,
            shell=False,
            **self.data
        )
        return result

    def release(self):
        with self.paths.repo.chdir:
            self.deploy()
        self



from typing import TypeVar, Generic

T = TypeVar("T")


class Inherit(Generic[T]):
    """

    Runtime inheritance. Acts like a wrapper around an instantiated base class of type T, and allows overriding methods in subclasses like regular inheritance.

    """

    def __init__(self, parent: T):
        """

        Set parent

        """
        object.__setattr__(self, "_parent", parent)

    def __getattribute__(self, name):
        """

        Since regular attribute access checks own methods first, we don't need to do anything fancy to fall back to the parent when not implemented.

        """

        cls = object.__getattribute__(self, "__class__")
        attr = cls.__dict__.get(name, None)
        if attr is not None and hasattr(attr, "__get__"):
            # descriptor on self class, call normally
            return attr.__get__(self, cls)
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            parent = object.__getattribute__(self, "_parent")
            return getattr(parent, name)

    @property
    def inherit_parent(self) -> T:
        """

        Retrieves the inherit parent `_parent` attribute.

        """
        return object.__getattribute__(self, "_parent")

    @property
    def inherit_root(self):
        """

        Returns the root object in the inheritance hierarchy.

        """
        cur = object.__getattribute__(self, "_parent")
        while isinstance(cur, Inherit):
            cur = object.__getattribute__(cur, "_parent")
        return cur



import pygit2 as vcs
from functools import cached_property
from typing import Any

from fmtr.tools.inherit_tools import Inherit
from fmtr.tools.logging_tools import logger
from fmtr.tools.path_tools import Path


class Repository(vcs.Repository):
    """

    Repository subclass to add some project-specific functionality.

    """
    SSH_DIR = Path().home() / ".ssh"

    def __init__(self, path: Path, project: Any):
        super().__init__(str(path))
        self.project = project


    @cached_property
    def tags(self):
        return Tags(self)

    @property
    def origin(self):
        return self.remotes["origin"]

    @property
    def keypair(self):
        return vcs.Keypair("git", self.SSH_DIR / "id_rsa.pub", self.SSH_DIR / "id_rsa", passphrase=None)

    @property
    def callbacks(self):
        return vcs.RemoteCallbacks(credentials=self.keypair)

    @logger.instrument('Fetching from repo {self.origin.url}...')
    def fetch(self):
        specs = [
            "+refs/heads/*:refs/remotes/origin/*",
            "+refs/tags/*:refs/tags/*",
        ]

        return self.origin.fetch(specs, callbacks=self.callbacks)

    @logger.instrument('Pushing to repo {self.origin.url}...')
    def push(self):
        allowed_heads = {"refs/heads/main", "refs/heads/release"}
        specs = [
            f"{ref}:{ref}"
            for ref in self.references
            if ref in allowed_heads or ref.startswith("refs/tags/")
        ]

        return self.origin.push(specs, callbacks=self.callbacks)





class Tags(Inherit[Repository]):

    @property
    def new(self):
        return f"v{self.project.repo.data.new}"

    @property
    def current(self):
        return f"v{self.project.version}"

    def get_tags(self):
        for ref in self.references:
            path = Path(ref)
            if path.parent.name == 'tags':
                yield path.name

    @property
    def all(self):
        return set(self.get_tags())



from functools import cached_property

from fmtr.tools import environment_tools as env
from fmtr.tools.constants import Constants
from fmtr.tools.docker_tools import DockerClient
from fmtr.tools.infrastructure_tools.project import Project
from fmtr.tools.inherit_tools import Inherit
from fmtr.tools.iterator_tools import IndexList
from fmtr.tools.logging_tools import logger
from fmtr.tools.merging_tools import merge
from fmtr.tools.path_tools import Path


class Stack(Inherit[Project]):
    """
    
    Represents the Docker stack for a project. Manages different deployment targets including:

    - Development compose stack with docker-compose configuration
    - Build targets for dev and production environments

    """

    @cached_property
    def cls(self):
        return self.__class__

    @cached_property
    def token(self):
        return env.get(Constants.CONTAINER_INDEX_PUBLIC_TOKEN_KEY)

    @cached_property
    def channel(self):
        return self.__class__.__name__.lower()

    @cached_property
    def path_compose(self):
        return Path.temp() / f'{self.name}.yml'

    @cached_property
    def client(self):
        return DockerClient(host=f"ssh://{self.hostname}", compose_files=self.path_compose)

    @classmethod
    def get_all(self):
        return [Development]

    @cached_property
    def composes_all(self):
        return IndexList[Compose](cls(self) for cls in (Compose, ComposeDocumentDatabase,))

    @cached_property
    def compose_data(self):
        """

        Merge Compose data for all project services

        """
        data = []
        for service in [Compose.NAME] + self.services:
            compose = self.composes_all.NAME[service]
            data.append(compose.data)
        data = merge(*data)
        return data

    @cached_property
    def tags_image(self):
        return [
            f'{self.name}:{self.channel}-{self.extras_str}',
            f'{self.name}:{self.channel}-{self.extras_str}-{self.tag}'
        ]

    @cached_property
    def entrypoint(self):
        # f"{self.org}-{self.package}-{self.entrypoint}" todo allow entrypoint from self.paths.entrypoints.
        return self.name_dash

    @logger.instrument('Building image for project {self.name} on channel {self.channel}...')
    def build(self):
        """

        Builds a Docker image using the specified build arguments, tags, and contexts.

        """

        build_args = dict(
            NAME=self.name,
            ORG=self.org,
            PACKAGE=self.package,
            BASE=self.base,
            EXTRAS=self.extras_str,
            ENTRYPOINT=self.entrypoint,
            SCRIPTS=self.scripts_str,
        )



        contexts = dict(
            package=str(self.paths.repo)
        )

        for line in self.client.build(
                build_contexts=contexts,
                file="Dockerfile",
                context_path=self.paths.repo,
                build_args=build_args,
                tags=self.tags_image,
                target=self.channel,
                load=True,
                progress="plain",
                stream_logs=True,
        ):
            logger.info(line.rstrip())


class Development(Stack):
    """

    Represents the development environment stack with channel-specific configuration

    """

    def recreate(self):
        """

        Recreates a compose deployment

        """

        self.build()

        data = self.compose_data

        with logger.span(f'Writing compose file to "{self.path_compose}"'):
            self.path_compose.write_yaml(data)

        self.client.compose.up(
            detach=True,
            force_recreate=True,

        )


class ProductionPrivate(Stack):
    @cached_property
    def channel(self):
        return 'production'

    def push(self):
        pass


class ProductionPublic(ProductionPrivate):

    @cached_property
    def tags_public(self):
        return [f'{Constants.ORG_NAME}/{self.name}:latest', f'{Constants.ORG_NAME}/{self.name}:{self.tag}']

    @cached_property
    def tags_image(self):
        tags = super().tags_image
        tags += self.tags_public
        return tags

    def push(self):
        self.client.login(username=Constants.ORG_NAME, password=self.token)
        for tag in self.tags_public:
            with logger.span(f'Pushing image "{tag}"'):
                for tag, line_bytes in self.client.push(tag, stream_logs=True):
                    line = line_bytes.decode().rstrip()
                    logger.info(line.rstrip())

        self

class Compose(Inherit[Stack]):
    """

    Compose file representation

    """
    NAME = 'base'

    @property
    def data(self):
        data = dict(
            name=f"{self.name_dash}",
            services=dict(
                interpreter=dict(
                    image=f"{self.name}:{self.channel}-{self.extras_str}",
                    restart="unless-stopped",
                    container_name=f"{self.name}",
                    hostname=f"{self.name_dash}-{self.channel}",
                    env_file=[
                        "/opt/dev/repo/env",
                    ],
                    environment=dict(
                        # DISPLAY=f"{self.display}",
                    ),
                    volumes=[
                        "dev:/opt/dev/repo",
                        "ssh:/home/user/.ssh",
                        "/opt/dev/data:/opt/dev/data",
                        "/home/user/.Xauthority:/home/user/.Xauthority:ro",
                        "/tmp/.X11-unix:/tmp/.X11-unix",
                    ],
                    ports=[
                        f"{2200 + self.port}:22",
                        f"{8000 + self.port}:8080",
                        f"{8100 + self.port}:8180",
                    ],
                    user="1000:1000",
                ),
            ),
            secrets=dict(
                hf_token=dict(
                    environment="HF_TOKEN",
                ),
            ),
            volumes=dict(
                dev=dict(
                    name=f"{self.name}",
                ),
                ssh=dict(
                    name="ssh",
                    external=True,
                ),
            ),
        )
        return data


class ComposeDocumentDatabase(Compose):
    """

    Compose file representation for document database service

    """
    NAME = 'db.document'

    @property
    def data(self):
        data = dict(
            # todo
        )
        return data
